@@ -29,8 +29,10 @@ using System;
 using System.Collections;
 using System.Diagnostics;
 using System.IO;
+using System.Xml;
+using System.Xml.Serialization;
 
-using BU = Beagle.Util;
+using Beagle.Util;
 
 namespace Beagle {
     
@@ -55,17 +57,14 @@ namespace Beagle {
 		// server-side copy of the Hit.  It is always null
 		// on the client-side.
 		private object sourceObject = null;
+		private string source_object_name = null;
 
 		// High scores imply greater relevance.
 		private double scoreRaw = 0.0;
 		private double scoreMultiplier = 1.0;
 
-		private Hashtable properties = new Hashtable (new CaseInsensitiveHashCodeProvider (), 
-							      new CaseInsensitiveComparer ());
-
-		private Hashtable data = new Hashtable (new CaseInsensitiveHashCodeProvider (), 
-							new CaseInsensitiveComparer ());
-
+		private ArrayList properties = new ArrayList ();
+		private bool sorted = true;
 
 		private enum SpecialType {
 			Unknown,
@@ -83,116 +82,80 @@ namespace Beagle {
 
 		//////////////////////////
 
-		public void WriteAsBinary (BinaryWriter writer)
-		{
-			writer.Write (BU.StringFu.DateTimeToString (Timestamp));
-			writer.Write (id);
-			writer.Write (BU.UriFu.UriToSerializableString (uri));
-			writer.Write (type);
-			writer.Write (mimeType == null ? "" : mimeType);
-			writer.Write (source);
-			writer.Write (scoreRaw);
-			writer.Write (scoreMultiplier);
-
-			writer.Write (properties.Count);
-			foreach (string key in properties.Keys) {
-				string value = (string) properties[key];
-				writer.Write (key);
-				writer.Write (value);
-			}
-
-			writer.Write (data.Count);
-			foreach (string key in data.Keys) {
-				byte[] value = (byte[]) data[key];
-				writer.Write (key);
-				writer.Write (value.Length);
-				writer.Write (value);
-			}
-		}
-
-		public static Hit ReadAsBinary (BinaryReader reader)
-		{
-			Hit hit = new Hit ();
-			
-			hit.Timestamp = BU.StringFu.StringToDateTime (reader.ReadString ());
-			hit.id = reader.ReadInt32 ();
-			hit.uri = BU.UriFu.UriStringToUri (reader.ReadString ());
-			hit.type = reader.ReadString ();
-			hit.mimeType = reader.ReadString ();
-			if (hit.mimeType == "")
-				hit.mimeType = null;
-			hit.source = reader.ReadString ();
-			hit.scoreRaw = reader.ReadDouble ();
-			hit.scoreMultiplier = reader.ReadDouble ();
-
-			int numProps = reader.ReadInt32 ();
-			for (int i = 0; i < numProps; i++) {
-				string key = reader.ReadString ();
-				string value = reader.ReadString ();
-
-				hit[key] = value;
-			}
-
-			int numData = reader.ReadInt32 ();
-			for (int i = 0; i < numData; i++) {
-				string key = reader.ReadString ();
-				int size = reader.ReadInt32 ();
-				byte[] data = reader.ReadBytes (size);
-
-				hit.SetData (key, data);
-			}
-
-			return hit;
-		}
-		
+		[XmlAttribute]
 		public int Id {
 			get { return id; }
 			set { id = value; }
 		}
 
+		[XmlIgnore]
 		public Uri Uri {
 			get { return uri; }
 			set { uri = value; }
 		}
 
+		[XmlAttribute ("Uri")]
+		public string UriAsString {
+			get {
+				return UriFu.UriToSerializableString (uri);
+			}
+
+			set {
+				uri = UriFu.UriStringToUri (value);
+			}
+		}
+
+		[XmlAttribute]
 		public string Type {
 			get { return type; }
 			set { type = value; }
 		}
 
+		[XmlAttribute]
 		public string MimeType {
 			get { return mimeType; }
 			set { mimeType = value; }
 		}
 	
+		[XmlAttribute]
 		public string Source {
 			get { return source; }
 			set { source = value; }
 		}
 
+		[XmlIgnore]
 		public object SourceObject {
 			get { return sourceObject; }
 			set { sourceObject = value; }
 		}
 
+		[XmlAttribute]
+		public string SourceObjectName {
+			get { return source_object_name; }
+			set { source_object_name = value; }
+		}
+
+		[XmlAttribute]
 		public double Score {
 			get { return scoreRaw * scoreMultiplier; }
 		}
 
+		[XmlAttribute]
 		public double ScoreRaw {
 			get { return scoreRaw; }
 			set { scoreRaw = value; }
 		}
 
+		[XmlAttribute]
 		public double ScoreMultiplier {
 			get { return scoreMultiplier; }
 			set { 
 				scoreMultiplier = value;
 				if (scoreMultiplier < 0) {
-					BU.Logger.Log.Warn ("Invalid ScoreMultiplier={0} for {1}", scoreMultiplier, Uri);
+					Logger.Log.Warn ("Invalid ScoreMultiplier={0} for {1}", scoreMultiplier, Uri);
 					scoreMultiplier = 0;
 				} else if (scoreMultiplier > 1) {
-					BU.Logger.Log.Warn ("Invalid ScoreMultiplier={0} for {1}", scoreMultiplier, Uri);
+					Logger.Log.Warn ("Invalid ScoreMultiplier={0} for {1}", scoreMultiplier, Uri);
 					scoreMultiplier = 1;
 
 				}
@@ -252,6 +215,7 @@ namespace Beagle {
 			get { return Path != null ? System.IO.Path.GetDirectoryName (Path) : null; }
 		}
 
+		[XmlIgnore]
 		public FileSystemInfo FileSystemInfo {
 			get {
 				if (IsFile)
@@ -263,6 +227,7 @@ namespace Beagle {
 			}
 		}
 
+		[XmlIgnore]
 		public FileInfo FileInfo {
 			get { 
 				if (fileInfo == null && IsFile)
@@ -271,6 +236,7 @@ namespace Beagle {
 			}
 		}
 
+		[XmlIgnore]
 		public DirectoryInfo DirectoryInfo {
 			get {
 				if (directoryInfo == null && IsDirectory)
@@ -281,52 +247,97 @@ namespace Beagle {
 
 		//////////////////////////
 
-		public IDictionary Properties {
-			get { return properties; }
+		[XmlArray]
+		[XmlArrayItem (ElementName="Property", Type=typeof (Property))]
+		public ArrayList Properties {
+			get {  return properties; }
 		}
 
-		public ICollection Keys {
-			get { return properties.Keys; }
+		public void AddProperty (Property prop)
+		{
+			if (sorted && properties.Count > 0) {
+				Property last_prop;
+				last_prop = properties [properties.Count - 1] as Property;
+				if (last_prop.CompareTo (prop) > 0) // i.e. last_prop > prop
+					sorted = false;
+			}
+
+			properties.Add (prop);
 		}
 
-		virtual public string this [string key] {
-			get { return (string) properties [key]; }
-			set { 
-				if (value == null || value == "") {
-					if (properties.Contains (key))
-						properties.Remove (key);
+		private bool FindProperty (string key, out int first, out int top)
+		{
+			// FIXME: Should use binary search on sorted property list
+			if (! sorted) {
+				properties.Sort ();
+				sorted = true;
+			}
+			
+			first = 0;
+			top = 0;
+
+			while  (first < properties.Count) {
+				Property prop;
+				prop = properties [first] as Property;
+				if (prop.Key == key)
+					break;
+				++first;
+			}
+
+			if (first >= properties.Count)
+				return false;
+
+			top = first + 1;
+			while (top < properties.Count) {
+				Property prop;
+				prop = properties [top] as Property;
+				if (prop.Key != key)
+					break;
+				++top;
+			}
+
+			return true;
+		}
+
+		private string this [string key] {
+			get {
+				int first, top;
+				if (! FindProperty (key, out first, out top))
+					return null;
+
+				if (top - first != 1) {
+					Logger.Log.Debug ("Accessed multi-property key with Hit's indexer.");
+					return null;
+				}
+
+				Property prop;
+				prop = properties [first] as Property;
+				return prop.Value;
+			}
+
+			set {
+				int first = 0, top = 0;
+
+				// If we've never heard of this property, add it.
+				if (! FindProperty (key, out first, out top)) {
+					AddProperty (Property.New (key, value));
 					return;
 				}
-				properties [key] = value as string;
+
+				// If it has appeared once before, clobber the existing
+				// value.  This emulates the previous (broken) semantics.
+
+				if (top - first == 1) {
+					properties [first] = Property.New (key, value);
+					return;
+				}
+				
+				// Otherwise throw an exception (which sort of sucks,
+				// but we don't really know what to do there)
+				throw new Exception (String.Format ("Attempt to re-set multi-property '{0}' via the indexer", key));
 			}
 		}
 
-		//////////////////////////
-
-		virtual public IDictionary Data {
-			get { return data; }
-		}
-
-		virtual public ICollection DataKeys {
-			get { return data.Keys; }
-		}
-
-		virtual public byte [] GetData (string key)
-		{
-			return (byte []) data [key];
-		}
-
-		virtual public void SetData (string key, byte [] blob)
-		{
-			if (blob == null) {
-				if (data.Contains (key))
-					data.Remove (key);
-				return;
-			}
-			data [key] = blob ;
-		}
-		
-
 		//////////////////////////
 
 		public override int GetHashCode ()
