@@ -29,11 +29,16 @@ using System.Diagnostics;
 using System.Globalization;
 using System.IO;
 using System.Runtime.InteropServices;
+using System.Text.RegularExpressions;
+using Mono.Posix;
 
 namespace Beagle.Util {
 
 	public class SystemInformation {
 
+		[DllImport ("libc", SetLastError=true)]
+		static extern int getloadavg (double[] loadavg, int nelem);
+
 		const double loadavg_poll_delay = 3;
 		static DateTime proc_loadavg_time  = DateTime.MinValue;
 		static double cached_loadavg_1min  = -1;
@@ -42,25 +47,21 @@ namespace Beagle.Util {
 
 		static private void CheckLoadAverage ()
 		{
-			// Only scan /proc/loadavg at most once every 10 seconds
+			// Only call getloadavg() at most once every 10 seconds
 			if ((DateTime.Now - proc_loadavg_time).TotalSeconds < loadavg_poll_delay)
 				return;
 
-			Stream stream = new FileStream ("/proc/loadavg",
-							FileMode.Open,
-							FileAccess.Read,
-							FileShare.Read);
-			TextReader reader = new StreamReader (stream);
+			double [] loadavg = new double [3];
+			int retval = getloadavg (loadavg, 3);
 
-			string line = reader.ReadLine ();
-			
-			reader.Close ();
-			stream.Close ();
+			if (retval == -1)
+				throw new IOException ("Could not get system load average: " + Syscall.strerror (Marshal.GetLastWin32Error ()));
+			else if (retval != 3)
+				throw new IOException ("Could not get system load average: getloadavg() returned an unexpected number of samples");
 
-			string [] parts = line.Split (' ');
-			cached_loadavg_1min = Double.Parse (parts [0], NumberFormatInfo.InvariantInfo);
-			cached_loadavg_5min = Double.Parse (parts [1], NumberFormatInfo.InvariantInfo);
-			cached_loadavg_15min = Double.Parse (parts [2], NumberFormatInfo.InvariantInfo);
+			cached_loadavg_1min  = loadavg [0];
+			cached_loadavg_5min  = loadavg [1];
+			cached_loadavg_15min = loadavg [2];
 
 			proc_loadavg_time = DateTime.Now;
 		}
@@ -117,12 +118,6 @@ namespace Beagle.Util {
 			if ((DateTime.Now - screensaver_time).TotalSeconds < screensaver_poll_delay)
 				return;
 
-			if (Environment.GetEnvironmentVariable ("EXERCISE_THE_DOG") != null) {
-				cached_screensaver_running = true;
-				cached_screensaver_idle_time = 24 * 60 * 60;
-				return;
-			}
-
 			ScreenSaverState state;
 			ScreenSaverKind kind;
 			ulong til_or_since = 0, idle = 0;
@@ -161,7 +156,7 @@ namespace Beagle.Util {
 
 		///////////////////////////////////////////////////////////////
 
-		const double acpi_poll_delay = 5;
+		const double acpi_poll_delay = 30;
 		const string proc_ac_state_filename = "/proc/acpi/ac_adapter/AC/state";
 		const string ac_present_string = "on-line";
 		static bool proc_ac_state_exists = true;
@@ -183,7 +178,7 @@ namespace Beagle.Util {
 			}
 
 			Stream stream = new FileStream (proc_ac_state_filename,
-							FileMode.Open,
+							System.IO.FileMode.Open,
 							FileAccess.Read,
 							FileShare.Read);
 			TextReader reader = new StreamReader (stream);
@@ -220,18 +215,106 @@ namespace Beagle.Util {
 			get { return get_vmrss (); }
 		}
 
+		static private int disk_stats_read_reqs;
+		static private int disk_stats_write_reqs;
+		static private int disk_stats_read_bytes;
+		static private int disk_stats_write_bytes;
+
+		static private DateTime disk_stats_time = DateTime.MinValue;
+		static private double disk_stats_delay = 1.0;
+
+		static private uint major, minor;
+
+		// Update the disk statistics with data for block device on the (major,minor) pair.
+		static private void UpdateDiskStats ()
+		{
+			string buffer;
+
+			if (major == 0)
+				return;
+
+			// We only refresh the stats once per second
+			if ((DateTime.Now - disk_stats_time).TotalSeconds < disk_stats_delay)
+				return;
+
+			// Read in all of the disk stats
+			using (StreamReader stream = new StreamReader ("/proc/diskstats"))
+				buffer = stream.ReadToEnd ();
+
+			// Find our partition and parse it
+			const string REGEX = "[\\s]+{0}[\\s]+{1}[\\s]+[a-zA-Z0-9]+[\\s]+([0-9]+)[\\s]+([0-9]+)[\\s]+([0-9]+)[\\s]+([0-9]+)";
+			string regex = String.Format (REGEX, major, minor);
+			Regex r = new Regex (regex, RegexOptions.IgnoreCase | RegexOptions.Compiled);
+			for (System.Text.RegularExpressions.Match m = r.Match (buffer); m.Success; m = m.NextMatch ()) {
+				disk_stats_read_reqs = Convert.ToInt32 (m.Groups[1].ToString ());
+				disk_stats_read_bytes = Convert.ToInt32 (m.Groups[2].ToString ());
+				disk_stats_write_reqs = Convert.ToInt32 (m.Groups[3].ToString ());
+				disk_stats_write_bytes = Convert.ToInt32 (m.Groups[4].ToString ());
+			}
+
+			disk_stats_time = DateTime.Now;
+		}
+
+		// Get the (major,minor) pair for the block device from which the index is mounted.
+		static private void GetIndexDev ()
+		{
+			Mono.Posix.Stat stat = new Mono.Posix.Stat ();
+			if (Mono.Posix.Syscall.stat (PathFinder.StorageDir, out stat) != 0)
+				return;
+
+			major = (uint) stat.Device >> 8;
+			minor = (uint) stat.Device & 0xff;
+		}
+
+		static public int DiskStatsReadReqs {
+			get {
+				if (major == 0)
+					 GetIndexDev ();
+				UpdateDiskStats ();
+				return disk_stats_read_reqs;
+			}
+		}
+
+		static public int DiskStatsReadBytes {
+			get {
+				if (major == 0)
+					 GetIndexDev ();
+				UpdateDiskStats ();
+				return disk_stats_read_bytes;
+			}
+		}
+
+		static public int DiskStatsWriteReqs {
+			get {
+				if (major == 0)
+					 GetIndexDev ();
+				UpdateDiskStats ();
+				return disk_stats_write_reqs;
+			}
+		}
+
+		static public int DiskStatsWriteBytes {
+			get {
+				if (major == 0)
+					 GetIndexDev ();
+				UpdateDiskStats ();
+				return disk_stats_write_bytes;
+			}
+		}
+
 #if false
 		static void Main ()
 		{
 			Gtk.Application.Init ();
 			while (true) {
-				Console.WriteLine ("{0} {1} {2} {3} {4} {5} {6}",
+				Console.WriteLine ("{0} {1} {2} {3} {4} {5} {6} {7}",
 						   LoadAverageOneMinute,
 						   LoadAverageFiveMinute,
 						   LoadAverageFifteenMinute,
 						   ScreenSaverRunning,
 						   InputIdleTime,
 						   UsingBattery,
+						   DiskStatsReadReqs,
 						   VmSize);
 				System.Threading.Thread.Sleep (1000);
 			}
