@@ -29,19 +29,20 @@ using System.Collections;
 using System.Text.RegularExpressions;
 using System.IO;
 
+using Beagle.Util;
+
 namespace Beagle.Daemon.FileSystemQueryable {
 
 	public class FileNameFilter {
 		
 		static string home_dir;
-		Hashtable ignore_cache = new Hashtable ();
+		private FileSystemModel model = null;
 
 		static FileNameFilter ()
 		{
-			home_dir = Environment.GetEnvironmentVariable ("HOME");
+			home_dir = PathFinder.HomeDir;
 		}
 
-
 		private class Pattern {
 			private string exactMatch;
 			private string prefix;
@@ -95,36 +96,16 @@ namespace Beagle.Daemon.FileSystemQueryable {
 			}
 		}
 
-		static ArrayList LoadPatterns (string filename)
-		{
-			if (! File.Exists (filename))
-				return null;
-			
-			ArrayList array = new ArrayList ();
-			StreamReader sr = new StreamReader (filename);
-			string line;
-			while ((line = sr.ReadLine ()) != null) {
-				line = line.Trim ();
-				if (line.Length > 0) {
-					Pattern pattern = new Pattern (line);
-					array.Add (pattern);
-				}
-			}
-			
-			return array;
-		}
-
 		//////////////////////////////////////////////////////////////////////
 		
-		private ArrayList defaultPatternsToIgnore;
+		private Hashtable patterns_to_ignore;
 
 		private void SetupDefaultPatternsToIgnore ()
 		{
-			defaultPatternsToIgnore = new ArrayList ();
-
 			// Add our default skip patterns.
 			// FIXME: This probably shouldn't be hard-wired.  Or should it?
-			AddDefaultPatternToIgnore (".*",
+			AddPatternToIgnore (new string [] {
+						   ".*",
 						   "*~",
 						   "#*#",
 						   "*.o",
@@ -155,6 +136,7 @@ namespace Beagle.Daemon.FileSystemQueryable {
 						   "Makefile.am",
 						   "Makefile.in",
 						   "CVS",
+						   "SCCS",
 						   // Garbage generated by the autotools
 						   "conftest",
 						   "confdefs.h",
@@ -162,108 +144,32 @@ namespace Beagle.Daemon.FileSystemQueryable {
 						   "confstat*",
 						   "/conf[0-9]+.sh/",
 						   "/conf[0-9]+.file/"
-				);
-			
-			// Read the ~/.neverindex file, which contains patterns
-			// for files that should always be ignored.
-			string neverindex = Path.Combine (home_dir, ".neverindex");
-			ArrayList patterns = LoadPatterns (neverindex);
-			if (patterns != null) {
-				foreach (Pattern pattern in patterns)
-					defaultPatternsToIgnore.Add (pattern);
-			}
+			});
 		}
 
-		public void AddDefaultPatternToIgnore (string pattern)
+		public void AddPatternToIgnore (string pattern)
 		{
-			defaultPatternsToIgnore.Add (new Pattern (pattern));
+			if (! patterns_to_ignore.ContainsKey (pattern))
+				patterns_to_ignore.Add (pattern, new Pattern (pattern));
 		}
 		
-		public void AddDefaultPatternToIgnore (params string[] patterns)
+		public void AddPatternToIgnore (IEnumerable patterns)
 		{
 			foreach (string pattern in patterns)
-				AddDefaultPatternToIgnore (pattern);
+				AddPatternToIgnore (pattern);
 		}
 		
 		/////////////////////////////////////////////////////////////
 
-		public FileNameFilter ()
+		public FileNameFilter (FileSystemModel model)
 		{
+			this.model = model;
+			patterns_to_ignore = new Hashtable ();
 			SetupDefaultPatternsToIgnore ();
 		}
 
 		/////////////////////////////////////////////////////////////
 
-		private class PerDirectoryInfo {
-			string dir;
-			DateTime noindexTimestamp;
-			DateTime noindexLastCheck;
-			ArrayList patternsToIgnore;
-
-			public PerDirectoryInfo (string _dir)
-			{
-				dir = _dir;
-				ConditionallyLoad ();
-			}
-
-			// If the directory's .noindex file has changed, load it.
-			// FIXME: This should be done with inotify instead of using
-			// this sort of seek-happy polling.
-			private void ConditionallyLoad ()
-			{
-				// Only re-check the .noindex file every 11 seconds.
-				DateTime now = DateTime.Now;
-				if ((now - noindexLastCheck).TotalSeconds < 11)
-					return;
-				noindexLastCheck = now;
-
-				string noindex = Path.Combine (dir, ".noindex");
-
-				if (File.Exists (noindex)) {
-					DateTime timestamp = File.GetLastWriteTime (noindex);
-					if (timestamp != noindexTimestamp) {
-						patternsToIgnore = LoadPatterns (noindex);
-						noindexTimestamp = timestamp;
-					}
-				} else {
-					patternsToIgnore = null;
-				}
-			}
-
-			public bool Ignore (string name)
-			{
-				ConditionallyLoad ();
-				if (patternsToIgnore == null)
-					return false;
-				if (patternsToIgnore.Count == 0) // i.e. IgnoreAll
-					return true;
-				foreach (Pattern pattern in patternsToIgnore)
-					if (pattern.IsMatch (name))
-						return true;
-
-				return false;
-			}
-		}
-
-		
-		private Hashtable perDirectoryCache = new Hashtable ();
-
-		private PerDirectoryInfo GetPerDirectoryInfo (string dir)
-		{
-			if (dir == null)
-				return null;
-
-			PerDirectoryInfo info;
-			lock (perDirectoryCache) {
-				info = perDirectoryCache [dir] as PerDirectoryInfo;
-				if (info == null) {
-					info = new PerDirectoryInfo (dir);
-					perDirectoryCache [dir] = info;
-				}
-			}
-			return info;
-		}
-
 		// FIXME: We could make this more efficient by storing more information.
 		// In particular, if ~/foo is an IgnoreAll directory, we could store
 		// that info in the PerDirectoryInfo of subdir ~/foo/bar so that
@@ -272,36 +178,22 @@ namespace Beagle.Daemon.FileSystemQueryable {
 		{
 			if (! Path.IsPathRooted (path))
 				path = Path.GetFullPath (path);
-			
-			if (ignore_cache.Contains (path))
-				return (bool) ignore_cache [path];
 
-			if (path == home_dir) {
-				ignore_cache [path] = false;
-				return false;
-			}
+			foreach (FileSystemModel.Directory root in model.Roots)
+				if (root.FullName == path)
+					return false;
 
 			string name = Path.GetFileName (path);
 
-			foreach (Pattern pattern in defaultPatternsToIgnore)
-				if (pattern.IsMatch (name)) {
-					ignore_cache [path] = true;
+			foreach (Pattern pattern in patterns_to_ignore.Values)
+				if (pattern.IsMatch (name))
 					return true;
-				}
 
 			string dir = Path.GetDirectoryName (path);
-			PerDirectoryInfo perDir = GetPerDirectoryInfo (dir);
-
-			if (perDir == null || perDir.Ignore (name)) {
-				ignore_cache [path] = true;
-				return true;
-			}
 
 			// A file should be ignored if any of its parent directories
 			// is ignored.
-			bool rv = Ignore (dir);
-			ignore_cache [path] = rv;
-			return rv;
+			return Ignore (dir);
 		}
 
 		public bool Ignore (FileSystemInfo info)
